/*! ********************************************
 * SynthData.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
public final class SynthData
{
// Instance ---
	static private var _instance:SynthData = null;
	static internal function getInstance():SynthData{ return (_instance)? _instance : _instance = new SynthData(); }
	static public function disposeInstance():void{ _instance = null; }
	public function SynthData():void{}
// const ---
	// Volume, Pan
	// Log2 Volume Table
	static internal const LOG2VT_DYN_BITS     :int = 16;                    // const 16    96dB dynamic range (FM TL, 16Bit PCM)
	static internal const LOG2VT_DYN          :int = 1 << LOG2VT_DYN_BITS;  //             1bit per 6dBv
	static internal const LOG2VT_DYN_RES_BITS :int = 8;
	static internal const LOG2VT_DYN_RES      :int = 1 << LOG2VT_DYN_RES_BITS;
	static internal const LOG2VT_SIZE         :int = LOG2VT_DYN_BITS << LOG2VT_DYN_RES_BITS + 1;
	// Pow2 Table
	static internal const POW2T_OCTV_BITS :int = 13;  // const 13
	static internal const POW2T_OCTV_RES  :int = 1 << POW2T_OCTV_BITS;  // POW2T_OCTV_RES == LOG2VT_SIZE
	static internal const POW2T_OCTV_MASK :int = POW2T_OCTV_RES - 1;
	// Envelope Generator
	static internal const EG_MAX_LEVEL_BITS :int = 4 + LOG2VT_DYN_RES_BITS;  // constant 12 (4 = 16 LOG2VT_DYN_BITS)
	static internal const EG_MAX_LEVEL      :int = 1 << EG_MAX_LEVEL_BITS;   // 4096 = 16 * 256
	static internal const EG_MAX_CNT_BITS   :int = 30;
	static internal const EG_MAX_CNT        :int = 1 << EG_MAX_CNT_BITS;
	static internal const EG_CNT_FP         :int = EG_MAX_CNT_BITS - EG_MAX_LEVEL_BITS;  // for EGLevelTable.  xbit fixed-point  (precision 1 / (1 << x))
	
	static internal const EG_ENV_DYN_BITS   :int = 13;  // 1 ~ 16
	static internal const EG_ENV_DYN        :int = 1 << EG_ENV_DYN_BITS;
	static internal const EG_CNT_ENV_ATK    :int = int(1.0 / EG_ENV_DYN * EG_MAX_LEVEL) << EG_CNT_FP;  // 0
	static internal const EG_CNT_ENV_BTM    :int = (EG_ENV_DYN_BITS * LOG2VT_DYN_RES) << EG_CNT_FP;
	static internal const EG_SUSTAIN_BTM    :int = EG_ENV_DYN_BITS * 8;
	
	static internal const EG_REFERENCE_TIME :Number = 16.0;                           // to 1/2 Volt (-6dB) Time [EG : Rate  0]
	static internal const EG_SOUND_OFF_TIME :Number = 1.0 / (EG_REFERENCE_TIME * 2.0);  // to 1/2 Volt (-6dB) Time [EG : Rate 34]
	// LFO
	static internal const LFO_MAX_FREQ             :Number = 32.0;  // Hz
	static internal const LFO_TIMER_INIT_VALUE     :int = 0x40000000;  // initial Value - Phase Update Timer
	static internal const LFO_WAVE_TABLE_SIZE_BITS :int = 8;
	static internal const LFO_WAVE_TABLE_SIZE      :int = 1 << LFO_WAVE_TABLE_SIZE_BITS;  // PM[0,2,4...],AM[1,3,5...]
	// Phase Modulation
	
// Tables ---
	// Volume, Pan
	private var _VOLUME_TABLE_0dB:Vector.<Number> = null;
	private var _VOLUME_TABLE_6dB:Vector.<Number> = null;
	private var _PAN_TABLE_SIN   :Vector.<Number> = null;
	private var _PAN_TABLE_BAL   :Vector.<Number> = null;
	// Log2 Volume Table
	private var _LOG2V_TABLE   :Vector.<Number> = null;
	private var _LOG2V_TABLE_I :Vector.<int> = null;
	// Pow2 Table
	private var _POW2_TABLE      :Vector.<int> = null;
	private var _POW2T_RS_TABLE  :Vector.<int> = null;
	private var _POW2T_KEY_TABLE :Vector.<int> = null;
	// Envelope Generator
	private var _EG_TOTAL_LEVEL_TABLE :Vector.<int> = null;
	private var _EG_LEVEL_TABLE_UP    :Vector.<int> = null;
	private var _EG_LEVEL_TABLE_DW    :Vector.<int> = null;
	
	private var _EG_KEY_TABLE      :Vector.<int> = null;
	private var _EG_POW2T_RS_TABLE :Vector.<int> = null;
	
	private var _EG_CNT_GATE_TABLE :Vector.<int> = null;
	// LFO
	private var _LFO_WAVE_TABLES :Vector.<Vector.<int>> = null;
	// Phase Modulation
	private var _MUL_TABLE      :Vector.<int> = null;

// getter, creator ---
// Volume, Pan ==================================================
	static internal function createVOLUME_TABLE(size:int, c:Number = 0.0, reverse:Boolean = false):Vector.<Number>
	{
		// Curve ( quadratic x^2 )
		// Fader |    dB | volt
		//  100% | - 0db | 1.0
		//   50% | -12db | 0.25
		//   25% | -24db | 0.0625
		var i:int, j:int = 1, _cnt:int = 1,
			_table:Vector.<Number> = new Vector.<Number>(size, true);
		if (reverse) { j = size; _cnt = -1; }
		for (i = 0; i < size; i++, j+=_cnt)_table[i] = Math.pow(j / size, 2.0) * c;
		return _table;
	}
	private function _createPAN_TABLE(type:int):Vector.<Number>
	{
		const _PIS2:Number = 3.141592653589793 / 2.0;
		var i:int, v:int, 
			_table:Vector.<Number> = new Vector.<Number>(257, true);
		switch (type)
		{
			// Panning Law (Balance Pan)
			// Fader         |    dB | volt
			//  100%         | - 0db | 1.0
			//   50%(Center) | - 0db | 1.0
			//   25%         | - 6db | 0.5
			case 1:
				for (i = 0; i < 64; i++)_table[i] = 1.0;
				for (i = 64, v = 0; i < 192; i++, v++)_table[i] = 1.0 - v / 64.0;
				for (i = 192; i < 257; i++)_table[i] = 0.0;
				break;
			// Panning Law (Combine Pan)
			// Fader         |    dB | volt
			//  100%         | - 0db | 1.0
			//   50%(Center) | - 6db | 0.5
			//   25%         | -12db | 0.25
			case 2:
				for (i = 0; i < 64; i++)_table[i] = 1.0;
				for (i = 64, v = 0; i < 192; i++, v++)_table[i] = (128-v) / 128.0;
				for (i = 192; i < 257; i++)_table[i] = 0.0;
				break;
			// Panning Law (Sine Curve)
			default:
				for (i = 0; i < 64; i++)_table[i] = 1.0;
				for (i = 64, v = 0; i < 192; i++, v++)_table[i] = Math.sin(((128 - v) / 128.0) * _PIS2);
				for (i = 192; i < 257; i++)_table[i] = 0.0;
				break;
		}
		return _table;
	}
	internal function getVOLUME_TABLE_0dB():Vector.<Number>{ return (_VOLUME_TABLE_0dB)? _VOLUME_TABLE_0dB : _VOLUME_TABLE_0dB = createVOLUME_TABLE(128, 1.0); } 
	internal function getVOLUME_TABLE_6dB():Vector.<Number>{ return (_VOLUME_TABLE_6dB)? _VOLUME_TABLE_6dB : _VOLUME_TABLE_6dB = createVOLUME_TABLE(128, 0.5011872336272722); } 
	internal function getPAN_TABLE_SIN():Vector.<Number>{ return (_PAN_TABLE_SIN)? _PAN_TABLE_SIN : _PAN_TABLE_SIN = _createPAN_TABLE(0); } 
	internal function getPAN_TABLE_BAL():Vector.<Number>{ return (_PAN_TABLE_BAL)? _PAN_TABLE_BAL : _PAN_TABLE_BAL = _createPAN_TABLE(1); } 
// Log2 Volume Table ==================================================
	static private const _LOG2VT_LOGE	:Number =  -(LOG2VT_DYN_RES * 1.442695040888963387); // -ln(2)
	static private const _LOG2VT_BOADER	:Number = 1.0 / LOG2VT_DYN;
	// in [-1.0 ~ 1.0]
	static internal function toLog2vtIndexValue(vol:Number):int
	{
		// Positive Value
		if      (vol >  _LOG2VT_BOADER) return  int(Math.log( vol) * _LOG2VT_LOGE + 0.5) << 1;
		// Negative Value
		else if (vol < -_LOG2VT_BOADER) return (int(Math.log(-vol) * _LOG2VT_LOGE + 0.5) << 1) + 1;  // +1 = Nega
		// 0 value
		else return LOG2VT_SIZE;
	}
	internal function getLOG2V_TABLE():Vector.<Number>
	{
		if (_LOG2V_TABLE) return _LOG2V_TABLE;
		const _dynB:Number = LOG2VT_DYN_BITS, _maxV:Number = LOG2VT_DYN, _res:Number = LOG2VT_DYN_RES;
		const _tableSize:int = (LOG2VT_SIZE + 2) * 4;  // Wave + EG(TL) + EG(DYN) + AM (+2 = safe)
		var _nv:Number, _table:Vector.<Number> = new Vector.<Number>(_tableSize, true);
		for (var i:int = 0, j:int = 0; i < LOG2VT_SIZE; j++)
		{
			_nv = Math.pow(2.0, _dynB - j / _res) / _maxV;
			_table[i++] =  _nv;
			_table[i++] = -_nv;
		}
		//for (i = _size; i < _tableSize; i++)_table[i] = 0.0;
		return _LOG2V_TABLE = _table;
	}
	internal function getLOG2V_TABLE_I():Vector.<int>
	{
		if (_LOG2V_TABLE_I) return _LOG2V_TABLE_I;
		const _dynB:Number = LOG2VT_DYN_BITS, _res:Number = LOG2VT_DYN_RES;
		const _tableSize:int = (LOG2VT_SIZE + 2) * 4;  // Wave + EG(TL) + EG(DYN) + AM (+2 = safe)
		var _iv:int, _table:Vector.<int> = new Vector.<int>(_tableSize, true);
		for (var i:int = 0, j:int = 0; i < LOG2VT_SIZE; j++)
		{
			_iv = int(Math.pow(2.0, _dynB - j / _res));
			_table[i++] =  _iv;
			_table[i++] = -_iv;
		}
		//for (i = _size; i < _tableSize; i++)_table[i] = 0;
		return _LOG2V_TABLE_I = _table;
	}
// Pow2 Table ==================================================
	static private const _POW2T_LOGE:Number = POW2T_OCTV_RES * 1.442695040888963387;
	// in [1.0 ~ inf]
	static internal function toPow2tIndexValue(val:Number):int
	{
		if (val <= 1.0) return 0;
		else return int(Math.log(val) * _POW2T_LOGE + 0.5);
	}
	internal function getPOW2_TABLE():Vector.<int>
	{
		if (_POW2_TABLE) return _POW2_TABLE;
		const _octvRes:Number = POW2T_OCTV_RES;
		var _table:Vector.<int> = new Vector.<int>(POW2T_OCTV_RES, true);
		for (var i:int = 0; i < POW2T_OCTV_RES; i++)_table[i] = int(Math.pow(2, 30.0 + (i / _octvRes)) + 0.5);
		return _POW2_TABLE = _table;
	}
	internal function getPOW2T_RS_TABLE():Vector.<int>
	{
		if (_POW2T_RS_TABLE) return _POW2T_RS_TABLE;
		var i:int, _table:Vector.<int> = new Vector.<int>(128, true);
		for (i =  0; i <  30; i++)_table[i] = 30 - i;
		for (i = 30; i < 128; i++)_table[i] = 0;  // Safe Field
		return _POW2T_RS_TABLE = _table;
	}
	internal function getPOW2T_KEY_TABLE():Vector.<int>
	{
		if (_POW2T_KEY_TABLE) return _POW2T_KEY_TABLE;
		const _octvRes:Number = POW2T_OCTV_RES;
		var _table:Vector.<int> = new Vector.<int>(128, true);
		for (var i:int = 0; i < 128; i++)_table[i] = int(i / 12.0 * _octvRes + 0.5);
		return _POW2T_KEY_TABLE = _table;
	}
// Envelope Generator ==================================================
	internal function getEG_TOTAL_LEVEL_TABLE():Vector.<int>
	{
		if (_EG_TOTAL_LEVEL_TABLE) return _EG_TOTAL_LEVEL_TABLE;
		// Ref - YM2151
			// D6 D5 D4 D3 D2 D1  D0
			// 48 24 12 6  3  1.5 0.75 [dB]
		const _step:int = LOG2VT_SIZE >> 7;
		var i:int, _table:Vector.<int> = new Vector.<int>(128*2, true);
		for (i =   0; i < 128; i++)_table[i] = i * _step;  // 64(0.75dB) = 8192/128
		for (i = 128; i < 256; i++)_table[i] = LOG2VT_SIZE;  // VeloCtrl
		return _EG_TOTAL_LEVEL_TABLE = _table;
	}
	internal function getEG_LEVEL_TABLE_DW():Vector.<int>
	{
		if (_EG_LEVEL_TABLE_DW) return _EG_LEVEL_TABLE_DW;
		var _table:Vector.<int> = new Vector.<int>(EG_MAX_LEVEL, true);
		for (var i:int = 0; i < EG_MAX_LEVEL; i++)_table[i] = i << 1;
		return _EG_LEVEL_TABLE_DW = _table;
	}
	internal function getEG_LEVEL_TABLE_UP():Vector.<int>
	{
		if (_EG_LEVEL_TABLE_UP) return _EG_LEVEL_TABLE_UP;
		const _size:Number = EG_MAX_LEVEL;
		var i:int, _table:Vector.<int> = new Vector.<int>(EG_MAX_LEVEL, true);
		for (i = 0; i < EG_MAX_LEVEL; i++)_table[i] = toLog2vtIndexValue(i / _size);
		return _EG_LEVEL_TABLE_UP = _table;
	}
	internal function getEG_KEY_TABLE():Vector.<int>
	{
		if (_EG_KEY_TABLE) return _EG_KEY_TABLE;
		const  _octvRes:Number = POW2T_OCTV_RES;
		var i:int, _b:int, _table:Vector.<int> = new Vector.<int>(256, true);
		for (i = 0; i < 128; i++)_table[i] = 0;
		for (i = 128, _b = 1; i < 256; i++, _b++)_table[i] = int(_b / 48.0 * _octvRes);
		return _EG_KEY_TABLE = _table;
	}
	internal function getEG_POW2T_RS_TABLE():Vector.<int>
	{
		if (_EG_POW2T_RS_TABLE) return _EG_POW2T_RS_TABLE;
		var i:int, _table:Vector.<int> = new Vector.<int>(128, true);
		for (i =  0; i <= EG_MAX_CNT_BITS; i++)_table[i] = 30 - i;
		for (i = EG_MAX_CNT_BITS; i < 128; i++)_table[i] = 31 - EG_MAX_CNT_BITS; // Safe Field (value >= 1)
		return _EG_POW2T_RS_TABLE = _table;
	}
	internal function getEG_CNT_GATE_TABLE():Vector.<int>
	{
		if (_EG_CNT_GATE_TABLE) return _EG_CNT_GATE_TABLE;
			// D6 D5 D4 D3 D2 D1  D0
			// 48 24 12 6  3  1.5 0.75 [dB]
		const _btm:int = EG_ENV_DYN_BITS * 8, 
		      _step:int = (EG_MAX_LEVEL << EG_CNT_FP) / 128;
		var i:int, _table:Vector.<int> = new Vector.<int>(129, true);
		for (i =    0; i < _btm; i++)_table[i] = i * _step;
		for (i = _btm; i <  129; i++)_table[i] = EG_CNT_ENV_BTM;
		return _EG_CNT_GATE_TABLE = _table;
	}
// LFO ==================================================
	static internal function lfoTimerCoef(sampleRate:Number, steps:int):int
	{
		var _tableSize:Number = LFO_WAVE_TABLE_SIZE / 2.0,
		// Max Frame Rate - LFO Phase Step
			_lfoMaxFrameRate:Number = sampleRate / LFO_MAX_FREQ / _tableSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		return int(LFO_TIMER_INIT_VALUE / steps / _lfoMaxFrameRate + 0.5);
	}
	internal function getLFO_WAVE_TABLES():Vector.<Vector.<int>>
	{
		if (_LFO_WAVE_TABLES) return _LFO_WAVE_TABLES;
		const _SIZ:int = LFO_WAVE_TABLE_SIZE, _HLF:int = _SIZ >> 1,
		      _QTR:int = _SIZ >> 2, _TRI:int = _HLF + _QTR;
		const _SIZ_N:Number = _SIZ, _HLF_N:Number = _SIZ_N / 2.0, _QTR_N:Number = _SIZ_N / 4.0;
		var i:int, _nv:Number, _iv:int,
			_table:Vector.<int>, _tables:Vector.<Vector.<int>> = new Vector.<Vector.<int>>(4, true);
			
	// Tri ---
		_tables[0] = _table = new Vector.<int>(_SIZ, true);
		// PM
		for (i = 0; i < _QTR; i += 2)
		{
			_iv = int(i / _QTR_N * POW2T_OCTV_RES);
			_table[       i] = _iv;
			_table[_QTR + i] = POW2T_OCTV_RES - _iv;
			_table[_HLF + i] = -_iv;
			_table[_TRI + i] = _iv - POW2T_OCTV_RES;
		}
		// AM
		for (i = 0; i < _HLF; i+=2)	
		{
			_iv = int(i / _HLF_N * LOG2VT_SIZE);
			_table[       i + 1] = _iv;
			_table[_HLF + i + 1] = LOG2VT_SIZE - _iv;
		}
	// Rect ---
		_tables[1] = _table = new Vector.<int>(_SIZ, true);
		for (i = 0; i < _HLF; i+=2)	
		{
			// PM	
			_table[       i] =  POW2T_OCTV_RES;
			_table[_HLF + i] = -POW2T_OCTV_RES;
			// AM
			_table[       i + 1] = 0;
			_table[_HLF + i + 1] = LOG2VT_SIZE;
		}
	// Saw ---
		_tables[2] = _table = new Vector.<int>(_SIZ, true);
		for (i = 0; i < _SIZ; i+=2)	
		{
			// PM	
			_table[i    ] = POW2T_OCTV_RES - int(i / _HLF_N * POW2T_OCTV_RES);
			// AM
			_table[i + 1] = int(i / _SIZ_N * LOG2VT_SIZE);
		}
	// Noise ---
		_tables[3] = _table = new Vector.<int>(_SIZ, true);
		for (i = 0; i < _SIZ; i+=2)	
		{
			_nv = Math.random();
			// AM
			_table[i + 1] = int(_nv * LOG2VT_SIZE);
			// PM
			_nv = _nv * 2.0 - 1.0;
			_table[i    ] = int(_nv * POW2T_OCTV_RES);
		}
		return _LFO_WAVE_TABLES = _tables;
	}
// Phase Modulation ==================================================
	internal function getMUL_TABLE():Vector.<int>
	{
		if (_MUL_TABLE) return _MUL_TABLE;
		const _octvRes:int = POW2T_OCTV_RES;
		var i:int, _table:Vector.<int> = new Vector.<int>(16, true);
		for (i = 0; i < 16; i++)_table[i] = (i-1) * _octvRes;
		return _MUL_TABLE = _table;
	}
}
}